
// emacs: this is -*- c++ -*-
//
//   @file    nnlo_interface.cxx        
//
//            grid filling routines for interface with 
//            the nnlo code from CITATION       
//  
//   Copyright (C) 2015 M.Sutton (sutt@cern.ch)    
//
//   $Id: nnlo_interface.h, v0.0  Thu 4 Jul 2015 10:25:18 CEST sutt $


#include <cstdlib>
#include <iostream>
#include <sstream>
#include <string>
#include <cmath>

#include <sys/stat.h>


#include "amconfig.h"

#ifdef HAVE_APPL
// applgrid headers
#include "appl_grid/appl_grid.h"
#include "appl_grid/lumi_pdf.h"
#endif

// nnlo header files
#include "nnlo_interface.h"
#include "nnlo_utils.h"
#include "nnlo_common.h"



 
// external hook functions 
extern void (*appl_initptr)();
extern void (*appl_fillptr)( const double& obs, 
			     const double& wt, 
			     const double& wt_full );
extern void (*appl_fillrefptr)( const double& obs, 
				const double& wt );
extern void (*appl_termptr)();


extern "C" void evolvepdf_(const double& x, const double& Q, double* xf); 
extern "C" double alphaspdf_(const double& Q);
  

/// Declare the input and output grids
static std::string _gridname = "nnlo.root";
  
static std::vector<appl::grid*> _grids;


// external hook functions
extern "C" bool setup_nnlo_() {
  std::cout << "NNLOJET INFO: setting up hook functions ..." << std::endl;

  appl_initptr       = nnlo::init;
  appl_fillptr       = nnlo::fill;
  appl_fillrefptr    = nnlo::fill_ref;
  appl_termptr       = nnlo::term;

  return true;
}



/// Initialise the hook functions
static bool _ready = setup_nnlo_();
  



/// banner
std::string nnlo::banner() {
    std::string version = PACKAGE_VERSION;
    std::string webpage = PACKAGE_URL;
    std::string banner  = "\n";
    banner += "*****************************************************************************\n";
    banner += "*   NNLOJET:        E.W.N.Glover, J.Pires [arXiv:.....]                     *\n";
    banner += "*   nnlo-bridge: E.W.N.Glover, C. Gwenlan, T.A.Morgan, M.R.Sutton    ...    *\n";
    banner += "*   APPLgrid:       T. Carli et al.    [arXiv:0911.2985]                    *\n";
    banner += "*   fastNLO:        D. Britzger K.Rabbertz, M. Wobisch. et al. [arXiv:...]  *\n";
    banner += "*****************************************************************************\n";
    banner += "\n";

    banner += "\n";
    banner += "Web page: " + webpage + "\n";
    banner += "\n";
    return banner;
}


bool first_run = true;


//
// Initialization
//
void nnlo::init() {

  std::cout << banner() << std::endl;
  

#ifdef HAVE_APPL
  if ( file_exists( _gridname ) ) { 
    // Open the existing grid
    
    std::cout << "nnlo::init() ready " << _ready << std::endl; 
    std::cout << "nnlo::init() reading grid " << _gridname << " ..." << std::endl;
    
    _grids.push_back(new appl::grid(_gridname));
    _grids.back()->optimise(true);
    
    first_run = false;

  }
  else {
    
    std::cout << "nnlo::init() booking grid from scratch with name " << _gridname << " ..." << std::endl;

    /// Create a new grid - these should be set from some values 
    /// from the NNLOJET code

    /// eta bins from eta0_a_gg5g0q_i1_n1_z2_q1_r001

    int Nobs      = 50;
    double obsmin = 0.0;
    double obsmax = 5;  //don't change this anymore !!!!!

    int    NQ2     = 15;
    double Q2min   = 10;
    //    double Q2max   = 10*14000*14000;
    double Q2max   = 2e6;
    double Q2order = 3;

    int    Nx     = 30;
    double xmin   = 0.00001;
    double xmax   = 0.9999999;
    double xorder = 5;


    int leading_order = inppar_.njets;
    int nloops        = inppar_.nloop;

    std::cout << " njets: " <<  inppar_.njets << "\tnloops " << inppar_.nloop << std::endl;
    
    /// needs to become process dependent and taken from the QCD code     
    std::string genpdf = "nnlo.config";
    
    std::cout << "creating grid " << genpdf << std::endl;
    
    //    lumi_pdf lumi( genpdf );
    //    lumi.write_latex( std::cout );

    //    double apramval=5.;
    //    appl::grid::transformvar(apramval);


    int nigrids = 0;

    if ( nloops == 0 ) nigrids = 0;
    if ( nloops == 1 ) nigrids = 2;
    if ( nloops == 2 ) nigrids = 0;

    _grids.push_back(new appl::grid(NQ2,  Q2min,  Q2max, Q2order,  
				    Nx,   xmin,   xmax,  xorder,
				    Nobs, obsmin, obsmax, 
				    genpdf, leading_order, nigrids ));

    // Use the reweighting function
    _grids.back()->reweight(true);

    // Add documentation
    _grids.back()->addDocumentation( banner() );
  }

#endif

#ifdef HAVE_FASTNLO
  
  std::cout << "fastNLO initialisation" << std::endl;

#endif

}







/// some naughty counters just to make things
/// easier for the time being

static int icount = 0;
static int i=0;

//
// fill the applgrid grid with the weights
//
void nnlo::fill( const double& obs, const double& wt, const double& wt_full  ) {


#ifdef HAVE_APPL

  std::cout << __FUNCTION__ << " APPLgrid fill() " << std::endl; 

  if ( _grids.size()==0 ) return;
#endif  

#ifdef HAVE_FASTNLO

  std::cout << __FUNCTION__ << " APPLgrid fill() " << std::endl; 
  
#endif

#if 0

  /// although this is applgrid code, it is not protected, since at 
  /// some point, both grids need this information 
   
  int Nsubproc = _grids.back()->subProcesses( inppar_.nloop );

  static std::vector<double> weight(Nsubproc); 

 
  double xi1 =  pdf_.xi1;
  double xi2 =  pdf_.xi2;

  double Q  =  scales_.rscale;
  double Q2 =  Q*Q;


  weight[0] = histonorm_.wthist * wt  ... ; 

  int as_order = inppar_.njets + inppar_.nloop;

  for ( int iorder=0 ; iorder<as_order ; iorder++ ) weight[0] *= 2*M_PI;

  //  if ( icount<1000 ) std::cout <<  histonorm_.wthist << std::endl;

  static bool first = true;


#if 0

  /// need to establish hos the parton kinematics is passed around ...
  /// then this can be 


    icount++;
  
    double _obs =  ... ;

    int fill_order = ...;
    
    

    if ( first_run ) { 

#ifdef HAVE_APPL
      _grids.back()->fill_phasespace(  pdf_.xi1, pdf_.xi2,  Q2, _obs,  &weight[0], fill_order );
#endif

#ifdef HAVE_FASTNLO

#endif

    }
    else  { 
 	
#ifdef HAVE_APPL 
	_grids.back()->fill_grid( pdf_.xi1, pdf_.xi2, Q2, _obs,  &weight[0], fill_order );
	fill_ref( _obs, wt_full*histonorm_.wthist );
#endif

#ifdef HAVE_FASTNLO

#endif

    }
    
  

#endif

#endif

}






//
// fill the applgrid reference histogram.
//
void nnlo::fill_ref( const double& obs, const double& wt ) {    

  //  if ( !first_run ) { 
  //    std::cout << "nnlo::fill_ref() called" << std::endl; 
  //    std::cout << "  weight:  " << wt << "\t obs " << obs << std::endl; 
  //  }

  _grids.back()->getReference()->Fill( obs, wt );
}





void Normalise(TH1D* h) 
{ 
  for ( int ibin=1 ; ibin<=h->GetNbinsX() ; ibin++ ) { 
    double width = h->GetBinLowEdge(ibin+1) - h->GetBinLowEdge(ibin);
    h->SetBinContent( ibin, h->GetBinContent(ibin)/width );
  }
  return;
}


//
// Write out the grids
//
void nnlo::term() { 

 
  std::cout << "nnlo::term() called" << std::endl; 
  
  std::cout << "  iterations: " << runinfo_.itmax2 << "\n" 
	    << "  shots:      " << runinfo_.nshot3 << "\n" 
	    << "  total:      " << runinfo_.itmax2*runinfo_.nshot3 
	    << " (icount " << icount << " )" << std::endl; 
  
  
  /// we need to scale by the bin width ???

  double nweights = runinfo_.itmax2*runinfo_.nshot2;

  //  double nweights = icount;

  ///  _grids.back()->getReference()->Scale(1/(conv*binwidth[nh]/nweights));

#ifdef HAVE_APPL

  std::cout << __FUNCTION__ << " APPLgrid term() " << std::endl; 

  _grids.back()->run() = nweights; 
  
  /// hmmm, we divide by the bin width after the convolution
  /// if the weights are *already* divided by the bin width, we will 
  /// need to scale *up* the weights in each bin by the bin width
  /// as a hack ( for fixed binning ) do that here ... 
 
  TH1D* h = _grids.back()->getReference();
  
  double width = 0;
  for ( int ibin=1 ; ibin<=h->GetNbinsX() ; ibin++ ) { 
    width = h->GetBinLowEdge(ibin+1) - h->GetBinLowEdge(ibin);
  }
  
  
  /// scale up the grid by the number of weights - vegas already 
  /// includes a 1/nshots term for the integration 
  /// the grid will always be stored "unweighted"
  /// NB: it *doesn't* scale by 1/niter2, so shouldn't scale by that
  (*_grids.back()) *= runinfo_.nshot2*width; 
  
  /// scale up the reference histogram, since it already is correctly 
  /// weighted after the filling, but we scale it down by the number 
  /// of weights when we write it out, so that the histogram in the 
  /// root file is the actual cross section 
  _grids.back()->getReference()->Scale( runinfo_.nshot2*nweights );

  Normalise( _grids.back()->getReference() );
  
  _grids.back()->Write( _gridname );

#endif

#ifdef HAVE_FASTNLO

  std::cout << __FUNCTION__ << " fastNLO term() " << std::endl; 

#endif

}
  
